# -*- coding=utf-8 -*-
from whoosh.analysis import RegexAnalyzer,LowercaseFilter,StopFilter,StemFilter
from whoosh.analysis import Tokenizer,Token 
from whoosh.lang.porter import stem

from yaha import Cuttor, SurnameCutting, SuffixCutting, get_dict, DICTS
import re

STOP_WORDS = None
def __init_stop_words():
    global STOP_WORDS
    stop_words = []
    for t,v in get_dict(DICTS.EXT_STOPWORD).iteritems():
        stop_words.append(t)
    for t,v in get_dict(DICTS.STOPWORD).iteritems():
        stop_words.append(t)
    for t,v in get_dict(DICTS.STOP_SENTENCE).iteritems():
        stop_words.append(t)
    STOP_WORDS = frozenset(stop_words)
__init_stop_words()

accepted_chars = re.compile(ur"[\u4E00-\u9FA5]+")

_cuttor = Cuttor()
_cuttor.set_stage1_regex(re.compile('(\d+)|([a-zA-Z]+)', re.I|re.U))
_cuttor.add_stage(SurnameCutting())
_cuttor.add_stage(SuffixCutting())

class ChineseTokenizer(Tokenizer):
    def __call__(self,text,**kargs):
        words = _cuttor.tokenize(text, search=True)
        token  = Token()
        for (w,start_pos,stop_pos) in words:
            if not accepted_chars.match(w):
                if len(w)>1:
                    pass
                else:
                    continue
            token.original = token.text = w
            token.pos = start_pos
            token.startchar = start_pos
            token.endchar = stop_pos
            yield token

def ChineseAnalyzer(stoplist=STOP_WORDS,minsize=1,stemfn=stem,cachesize=50000):
    return ChineseTokenizer()|LowercaseFilter()|StopFilter(stoplist=stoplist,minsize=minsize)\
                                        |StemFilter(stemfn=stemfn, ignore=None,cachesize=cachesize)



# -*- coding=utf-8 -*-
import sys
import os, codecs, re, math
import collections
import threading
from yaha import BaseCuttor, WordBase, get_dict, DICTS

max_word_len = 5 
entropy_threshold = 1 
max_to_flush = 10000

class Word(WordBase):
    def __init__(self, id):
        super(WordBase, self).__init__()

        self.process_freq = 1
        self.total_freq = 1
        self.valid = 0
        self.process_ps = 0.0
        self.id = id
        self.l_len = 0
        self.r_len = 0
        self.l = collections.Counter()
        self.r = collections.Counter()
        self.base_freq = 0
        self.base_ps = 0.0
        self.curr_ps = 0.0

    def add(self):
        self.process_freq += 1
        self.total_freq += 1

    def add_l(self, word):
        if word in self.l:
            self.l[word] += 1
        else:
            self.l[word] = 1
        self.l_len += 1
    
    def add_r(self, word):
        if word in self.r:
            self.r[word] += 1
        else:
            self.r[word] = 1
        self.r_len += 1

    def reset(self, id):
        self.process_freq = 1
        self.id = id
        self.l_len = 0
        self.r_len = 0
        self.l = collections.Counter()
        self.r = collections.Counter()

# TODO has a better implement?
# How to add fields to Objects dynamically ?
MODIFY_LOCK = threading.RLock()
MODIFY_INIT = False
def modify_wordbase(word):
        word.process_freq = 1
        word.total_freq = 1
        word.valid = 0
        word.process_ps = 0.0
        word.id = id
        word.l_len = 0
        word.r_len = 0
        word.l = collections.Counter()
        word.r = collections.Counter()
        word.curr_ps = word.base_ps
        word.add = Word.add
        word.add_l = Word.add_l
        word.add_r = Word.add_r
        word.reset = Word.reset

def get_modified_dict():
    global MODIFY_INIT
    dict = get_dict(DICTS.MAIN)
    if MODIFY_INIT:
        return dict
    with MODIFY_LOCK:
        if MODIFY_INIT:
            return dict
        for word in dict:
            modify_wordbase(word)
        MODIFY_INIT = True
    return dict

def info_entropy(words, total):
    result = 0 
    for word, cnt in words.iteritems():
        p = float(cnt) / total
        result -= p * math.log(p)
    return result

class Process(object):
    def __init__(self, id):
        self.id = id
        self.words = []
        self.cache_lines = []
    
    def add_words(self, word):
        self.words.append(word)

    def do_sentence(self, sentence, word_dict):
        l = len(sentence)
        wl = min(l, max_word_len)
        self.cache_lines.append(sentence)
        for i in xrange(1, wl + 1): 
            for j in xrange(0, l - i + 1): 
                if j == 0:
                    if j < l-i:
                        word_dict.add_word_r(sentence[j:j+i], sentence[j+i])
                    else:
                        word_dict.add_word(sentence[j:j+i])
                else:
                    if j < l-i:
                        word_dict.add_word_lr(sentence[j:j + i], sentence[j-1], sentence[j+i])
                    else:
                        word_dict.add_word_l(sentence[j:j+i], sentence[j-1])

    def calc(self, word_dict):
        # calc all ps first
        for word in self.words:
            this_word = word_dict.get_word(word)
            this_word.process_ps = float(this_word.process_freq)/word_dict.process_total
    
        # then calc the ps around the word
        for word in self.words:
            this_word = word_dict.get_word(word)
            if len(word) > 1:
                p = 0
                for i in xrange(1, len(word)):
                    t = word_dict.ps(word[0:i]) * word_dict.ps(word[i:])
                    p = max(p, t)
                if p > 0 and this_word.process_freq >= 3 and this_word.process_ps / p > 100:
                    if this_word.l_len > 0 and info_entropy(this_word.l, this_word.l_len) < entropy_threshold:
                        continue
                    if this_word.r_len > 0 and info_entropy(this_word.r, this_word.r_len) < entropy_threshold:
                        continue
                    this_word.valid += 1
                    this_word.curr_ps = math.log(float(this_word.total_freq+this_word.base_freq)/float(word_dict.base_total+word_dict.total/word_dict.id))

class WordDict(BaseCuttor):

    def __init__(self, new_dict=True):
        super(WordDict, self).__init__()

        self.dict = {}
        self.total = 0
        self.base_total = 0
        self.id = 0
        self.process_total = 0
        self.current_line = 0
        
        self.WORD_MAX = 5
        
        '''with codecs.open(dict_file, "r", "utf-8") as file:
            for line in file:
                tokens = line.split(" ")
                word = tokens[0].strip()
                if len(tokens) >= 2:
                    this_word = Word(0)
                    freq = int(tokens[1].strip())
                    this_word.base_freq = freq
                    self.dict[word] = this_word
                    self.base_total += freq
        
        #normalize
        for word, term in self.dict.iteritems():
            term.base_ps = math.log(float(term.base_freq)/self.base_total)
            term.curr_ps = term.base_ps'''

        if not new_dict:
            # TODO for getting dict from MAIN_DICT
            self.dict = get_modified_dict()

        self.new_process()

    def add_user_dict(self, filename):
        with codecs.open(filename, "r", "utf-8") as file:
            for line in file:
                tokens = line.split(" ")
                word = tokens[0].strip()
                if len(tokens) >= 2:
                    freq = int(tokens[1].strip())
                    if word in self.dict:
                        this_word = self.dict[word]
                        this_word.base_freq += freq
                        self.base_total += freq
                    else:
                        this_word = Word(0)
                        this_word.base_freq = freq
                        self.dict[word] = this_word
                        self.base_total += freq
        #normalize
        for word, term in self.dict.iteritems():
            term.base_ps = math.log(float(term.base_freq)/self.base_total)
            term.curr_ps = term.base_ps
    
    def exist(self, word):
        if word not in self.dict:
            return False
        this_word = self.dict[word]
        return (this_word.curr_ps < 0.0) or (this_word.valid > self.id/2)

    def get_prob(self, word):
        if word in self.dict:
            return self.dict[word].curr_ps
        else:
            return 0.0
    
    def new_process(self):
        self.id += 1
        self.process = Process(self.id)
        self.process_total = 0
        return self.process
    
    def add_word(self, word):
        this_word = None
        if word in self.dict:
            this_word = self.dict[word]
            if self.id == this_word.id:
                this_word.add()
            else:
                this_word.reset(self.id)
                self.process.add_words(word)
        else:
            this_word = Word(self.id)
            self.dict[word] = this_word
            self.process.add_words(word)
        self.process_total += 1
        self.total += 1
        return this_word

    def learn(self, sentence):
        for s,need_cut in self.cut_to_sentence(sentence):
            if not need_cut:
                continue
            self.process.do_sentence(s, self)
            self.current_line += 1
            if self.current_line > max_to_flush:
                self.process.calc(self)
                self.new_process()
                self.current_line = 0

    def learn_flush(self):
        self.process.calc(self)
        self.new_process()
        self.current_line = 0

    def cut_and_learn(self, sentence):
        self.learn(sentence)
        self.cut(sentence)

    def add_word_l(self, word, l):
        w = self.add_word(word)
        w.add_l(l)
    
    def add_word_r(self, word, r):
        w = self.add_word(word)
        w.add_r(r)
    
    def add_word_lr(self, word, l, r):
        w = self.add_word(word)
        w.add_l(l)
        w.add_r(r)

    def ps(self, word):
        if word in self.dict and self.dict[word].id == self.id:
            return self.dict[word].process_ps
        else:
            return 0.0

    def get_word(self, word):
        return self.dict[word]

    def save_to_file(self, filename):
        word_dict = self
        final_words = [] 
        for word, term in word_dict.dict.iteritems():
            #if term.valid > word_dict.id/2 and term.base_freq == 0:
            # Use this to save more word
            if term.valid > 0 and term.base_freq == 0:
                final_words.append(word)

        final_words.sort(cmp = lambda x, y: cmp(word_dict.get_word(y).total_freq, word_dict.get_word(x).total_freq))
        
        with codecs.open(filename, 'w', 'utf-8') as file:
            for word in final_words:
                v = word_dict.get_word(word).total_freq
                file.write("%s %d\n" % (word,v))



# -*- coding: utf-8 -*-
#
#  prioritydictionary.py
#  
#  Copyright 2002 David Eppstein, UC Irvine
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
# http://code.activestate.com/recipes/117228/
# 
from __future__ import generators

class priorityDictionary(dict):
    def __init__(self):
        '''Initialize priorityDictionary by creating binary heap of pairs 
        (value,key).  Note that changing or removing a dict entry will not 
        remove the old pair from the heap until it is found by smallest() or
		until the heap is rebuilt.'''
        self.__heap = []
        dict.__init__(self)

    def smallest(self):
        '''Find smallest item after removing deleted items from heap.'''
        if len(self) == 0:
            raise IndexError, "smallest of empty priorityDictionary"
        heap = self.__heap
        while heap[0][1] not in self or self[heap[0][1]] != heap[0][0]:
            lastItem = heap.pop()
            insertionPoint = 0
            while 1:
                smallChild = 2*insertionPoint+1
                if smallChild+1 < len(heap) and \
                        heap[smallChild] > heap[smallChild+1]:
                    smallChild += 1
                if smallChild >= len(heap) or lastItem <= heap[smallChild]:
                    heap[insertionPoint] = lastItem
                    break
                heap[insertionPoint] = heap[smallChild]
                insertionPoint = smallChild
        return heap[0][1]
	
    def __iter__(self):
        '''Create destructive sorted iterator of priorityDictionary.'''
        def iterfn():
            while len(self) > 0:
                x = self.smallest()
                yield x
                del self[x]
        return iterfn()
	
    def __setitem__(self,key,val):
        '''Change value stored in dictionary and add corresponding pair to heap.  
        Rebuilds the heap if the number of deleted items grows too large, to 
        avoid memory leakage.'''
        dict.__setitem__(self,key,val)
        heap = self.__heap
        if len(heap) > 2 * len(self):
            self.__heap = [(v,k) for k,v in self.iteritems()]
            self.__heap.sort()  # builtin sort likely faster than O(n) heapify
        else:
            newPair = (val,key)
            insertionPoint = len(heap)
            heap.append(None)
            while insertionPoint > 0 and \
                    newPair < heap[(insertionPoint-1)//2]:
                heap[insertionPoint] = heap[(insertionPoint-1)//2]
                insertionPoint = (insertionPoint-1)//2
            heap[insertionPoint] = newPair
	
    def setdefault(self,key,val):
        '''Reimplement setdefault to call our customized __setitem__.'''
        if key not in self:
            self[key] = val
        return self[key]




# -*- coding=utf-8 -*-
from operator import itemgetter
from prioritydictionary import priorityDictionary

class Graph:
    INFINITY = 10000
    UNDEFINDED = None

    def __init__(self, n):
        self._data = {}
        self.N = n
        for i in xrange(0,n,1):
            self._data[i] = {}

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return (self._data)

    def __getitem__(self, node):
        if self._data.has_key(node):
            return self._data[node]
        else:
            return None
    
    def __iter__(self):
        return self._data.__iter__()
    
    def add_edge(self, node_from, node_to, cost=None):
        if not cost:
            cost = self.INFINITY

        self._data[node_to][node_from] = cost
        return
    
    def remove_edge(self, node_from, node_to, cost=None):
        if self._data[node_to].has_key(node_from):
            if not cost:
                cost = self._data[node_to][node_from]
                
                if cost == self.INFINITY:
                    return -1
                else:
                    self._data[node_to][node_from] = self.INFINITY
                    return cost
            elif self._data[node_to][node_from] == cost:
                self._data[node_to][node_from] = self.INFINITY
                
                return cost
            else:
                return -1
        else:
            return -1

def ksp_yen(graph, node_start, node_end, max_k=3):
    distances, previous = dp_graph(graph, node_start)
    #print 'distance=',distances
    #print 'previous=',previous
    
    A = [{'cost': distances[node_end], 
          'path': path(previous, node_start, node_end)}]
    B = []
    
    if not A[0]['path']: return A
    
    for k in range(1, max_k):
        for i in range(0, len(A[-1]['path']) - 1):
            node_spur = A[-1]['path'][i]
            path_root = A[-1]['path'][:i+1]
            
            edges_removed = []
            #print '\n A=', A
            for path_k in A:
                curr_path = path_k['path']
                if len(curr_path) > i and path_root == curr_path[:i+1]:
                    cost = graph.remove_edge(curr_path[i], curr_path[i+1])
                    #print 'remove ', cost, curr_path[i], curr_path[i+1]
                    if cost == -1:
                        continue
                    edges_removed.append([curr_path[i], curr_path[i+1], cost])
            
            path_spur = dp_graph(graph, node_spur, node_end)
            #print k,i,node_spur,node_end,'path_spur=',path_spur
            
            if path_spur['path']:
                path_total = path_root[:-1] + path_spur['path']
                dist_total = distances[node_spur] + path_spur['cost']
                potential_k = {'cost': dist_total, 'path': path_total}
            
                if not (potential_k in B):
                    B.append(potential_k)
            
            for edge in edges_removed:
                graph.add_edge(edge[0], edge[1], edge[2])
        
        if len(B):
            B = sorted(B, key=itemgetter('cost'))
            A.append(B[0])
            B.pop(0)
        else:
            break
    
    return A

def dp_graph(graph, node_start, node_end=None):
    N = graph.N
    distances = {} 
    previous = {}
    
    previous[node_start] = None
    for idx in xrange(0, node_start+1, 1):
        distances[idx] = 0.0

    for idx in xrange(node_start+1,N,1):
        Q = priorityDictionary()
        for x in graph[idx]:
            Q[x] = distances[x] + graph[idx][x]
        small = Q.smallest()
        if small < node_start:
            previous[idx] = node_start
        else:
            previous[idx] = Q.smallest()
        distances[idx] = Q[small]
    
    if node_end:
        return {'cost': distances[node_end], 
                'path': path(previous, node_start, node_end)}
    else:
        return (distances, previous)

def path(previous, node_start, node_end):
    route = []

    node_curr = node_end    
    while True:
        route.append(node_curr)
        if previous[node_curr] == node_start:
            route.append(node_start)
            break
        elif previous[node_curr] == Graph.UNDEFINDED:
            return []
        
        node_curr = previous[node_curr]
    
    route.reverse()
    return route



# -*- coding=utf-8 -*-
import sys
import os, codecs, re, math
import collections
import threading
import codecs
from math import sqrt
from yaha import DICTS, get_dict, Cuttor
from analyzer import ChineseAnalyzer

IDF_LOCK = threading.RLock()
IDF_INIT = False
idf_freq = {}
DEFAULT_IDF = 1.5

def __init_idf():
    global idf_freq, IDF_INIT
    if IDF_INIT:
        return
    with IDF_LOCK:
        if IDF_INIT:
            return
    idf_path = os.path.normpath( os.path.join( os.getcwd(), os.path.dirname(__file__) )  )
    # the idf.txt file is make by whoosh,
    # I will put this code to extra/idfmake.py 21/08/13 08:39:22
    idf_name = os.path.join(idf_path, "idf.txt")
    content = ''
    num_freqs = 0
    with codecs.open(idf_name, 'r', 'utf-8') as corpus_file:
        # Load number of documents.
        line = corpus_file.readline()
        num_docs = int(line.strip())

        for line in corpus_file:
            tokens = line.split(" ")
            term = tokens[0].strip()
            frequency = int(tokens[1].strip())
            idf_freq[term] = frequency
            num_freqs += float(frequency)

        for term,freq in idf_freq.iteritems():
            idf_freq[term] = math.log(float(1 + num_docs) / (1 + freq))

    IDF_INIT = True

__init_idf()

median_idf = sorted(idf_freq.values())[len(idf_freq)/2]

def extract_keywords(content, topk=18, cuttor=None):
    stopwords = get_dict(DICTS.STOPWORD)

    tmp_cuttor = None
    if cuttor:
        tmp_cuttor = cuttor
    else:
        tmp_cuttor = Cuttor()
        #support for number and english 21/08/13 08:43:23
        tmp_cuttor.set_stage1_regex(re.compile('(\d+)|([a-zA-Z]+)', re.I|re.U))

    words = tmp_cuttor.cut(content)
    freq = {}
    total = 0
    for word in words:
        if len(word.strip()) < 2:
            continue
        lower_word = word.lower()
        if stopwords.has_key(lower_word):
            continue
        #TODO only leave the 'n' word? 21/08/13 09:13:36
        if tmp_cuttor.exist(lower_word) and not tmp_cuttor.word_type(lower_word, 'n'):
            continue
        total += 1
        if word in freq:
            freq[lower_word] += 1
        else:
            freq[lower_word] = 1
    freq = [(k,v/total) for k,v in freq.iteritems()]
    tf_idf_list = [(v * idf_freq.get(k,median_idf),k) for k,v in freq]
    st_list = sorted(tf_idf_list, reverse=True)

    top_tuples = st_list[:topk]
    keys = [a[1] for a in top_tuples]
    return keys

def near_duplicate(content1, content2, cuttor=None):
    tmp_cuttor = None
    if cuttor:
        tmp_cuttor = cuttor
    else:
        tmp_cuttor = Cuttor()
        tmp_cuttor.set_stage1_regex(re.compile('(\d+)|([a-zA-Z]+)', re.I|re.U))
    file_words = {}
    stopwords = [get_dict(DICTS.STOP_SENTENCE), get_dict(DICTS.EXT_STOPWORD), get_dict(DICTS.STOPWORD)]

    seg_list = tmp_cuttor.cut(content1)
    for w in seg_list:
        is_drop = False
        lw = w.lower()
        for stopword in stopwords:
            if stopword.has_key(lw):
                is_drop = True
                break
            if is_drop:
                continue
            if lw not in file_words.keys():
                file_words[lw] = [1,0]
            else:
                file_words[lw][0] += 1
    
    seg_list = tmp_cuttor.cut(content2)
    for w in seg_list:
        is_drop = False
        lw = w.lower()
        for stopword in stopwords:
            if stopword.has_key(lw):
                is_drop = True
                break
            if is_drop:
                continue
            if lw not in file_words.keys():
                file_words[lw] = [0,1]
            else:
                file_words[lw][1] += 1

    sum_2 = 0
    sum_file1 = 0
    sum_file2 = 0
    for word in file_words.values():
        sum_2 += word[0]*word[1]
        sum_file1 += word[0]**2
        sum_file2 += word[1]**2

    rate = sum_2/(sqrt(sum_file1*sum_file2))
    return rate

def __search_word(sentences, word):
    for s in sentences:
        if s.find(word)>=0:
            return s
    return ''

def summarize1(original_text, summary_size = 8, cuttor = None):
    if cuttor:
        tmp_cuttor = cuttor
    else:
        tmp_cuttor = Cuttor()
        tmp_cuttor.set_stage1_regex(re.compile('(\d+)|([a-zA-Z]+)', re.I|re.U))

    words_sorted = extract_keywords(original_text, 16, cuttor)
    summary_set = {}
    sentences = []
    for s,need in tmp_cuttor.cut_to_sentence(original_text):
        if need:
            sentences.append(s)

    for word in words_sorted:
        matching_sentence = __search_word(sentences, word)
        if matching_sentence <> '':
            summary_set[matching_sentence] = 1
            if len(summary_set) >= summary_size:
                break
    summary = []
    for s in sentences:
        if s in summary_set:
            summary.append(s)
    return ', '.join(summary)+'.'

N = 80  # Number of words to consider
CLUSTER_THRESHOLD = 5  # Distance between words to consider
TOP_SENTENCES = 8  # Number of sentences to return for a "top n" summary

def __score_sentences(sentences, important_words, cuttor):
    scores = []
    sentence_idx = -1

    for s in [list(cuttor.cut(s)) for s in sentences]:
        sentence_idx += 1
        word_idx = []

        # For each word in the word list...
        for w in important_words:
            try:
                # Compute an index for where any important words occur in the sentence

                word_idx.append(s.index(w))
            except ValueError, e: # w not in this particular sentence
                pass

        word_idx.sort()

        # It is possible that some sentences may not contain any important words at all
        if len(word_idx)== 0: continue

        # Using the word index, compute clusters by using a max distance threshold
        # for any two consecutive words

        clusters = []
        cluster = [word_idx[0]]
        i = 1
        while i < len(word_idx):
            if word_idx[i] - word_idx[i - 1] < CLUSTER_THRESHOLD:
                cluster.append(word_idx[i])
            else:
                clusters.append(cluster[:])
                cluster = [word_idx[i]]
            i += 1
        clusters.append(cluster)

        # Score each cluster. The max score for any given cluster is the score 
        # for the sentence

        max_cluster_score = 0
        for c in clusters:
            significant_words_in_cluster = len(c)
            total_words_in_cluster = c[-1] - c[0] + 1
            score = 1.0 * significant_words_in_cluster \
                * significant_words_in_cluster / total_words_in_cluster

            if score > max_cluster_score:
                max_cluster_score = score

        scores.append((sentence_idx, score))
    return scores

def summarize2(txt, cuttor=None):
    if cuttor:
        tmp_cuttor = cuttor
    else:
        tmp_cuttor = Cuttor()
        tmp_cuttor.set_stage1_regex(re.compile('(\d+)|([a-zA-Z]+)', re.I|re.U))
    
    sentences = []
    #TODO do it better 21/08/13 12:36:08
    for s,need in tmp_cuttor.cut_to_sentence(txt):
        if need:
            sentences.append(s)
    normalized_sentences = [s.lower() for s in sentences]

    top_n_words = extract_keywords(txt, N, tmp_cuttor)
    scored_sentences = __score_sentences(normalized_sentences, top_n_words, tmp_cuttor)

    top_n_scored = sorted(scored_sentences, key=lambda s: s[1])[-TOP_SENTENCES:]
    top_n_scored = sorted(top_n_scored, key=lambda s: s[0])
    top_n_summary=[sentences[idx] for (idx, score) in top_n_scored]
    return ', '.join(top_n_summary) + '.'

#
def summarize3(txt, cuttor=None):
    # TODO how to do this better 21/08/13 13:07:22
    # You can replace this with "import numpy", and of cause you have to
    # install the lib numpy
    name = "numpy"
    numpy = __import__(name, fromlist=[])

    if cuttor:
        tmp_cuttor = cuttor
    else:
        tmp_cuttor = Cuttor()
        tmp_cuttor.set_stage1_regex(re.compile('(\d+)|([a-zA-Z]+)', re.I|re.U))
    
    sentences = []
    #TODO do it better 21/08/13 12:36:08
    for s,need in tmp_cuttor.cut_to_sentence(txt):
        if need:
            sentences.append(s)
    normalized_sentences = [s.lower() for s in sentences]

    top_n_words = extract_keywords(txt, N, tmp_cuttor)
    scored_sentences = __score_sentences(normalized_sentences, top_n_words, tmp_cuttor)
    avg = numpy.mean([s[1] for s in scored_sentences])
    std = numpy.std([s[1] for s in scored_sentences])
    mean_scored = [(sent_idx, score) for (sent_idx, score) in scored_sentences
                   if score > avg + 0.5 * std]
    mean_scored_summary=[sentences[idx] for (idx, score) in mean_scored]
    return ', '.join(mean_scored_summary) + '.'



# -*- coding=utf-8 -*-
import sys, os.path, codecs, re, math
import collections
import threading
import tempfile
import random
import cPickle as pickle
import time

# ksp_dijkstra for default
from ksp_dijkstra import Graph, ksp_yen, quick_shortest
#note: a little error in ksp_dp
#from ksp_dp import Graph, ksp_yen

DICT_LOCK = threading.RLock()
DICT_INIT = False
class DICTS:
    LEN = 7
    MAIN,SURNAME,SUFFIX,EXT_STOPWORD,STOPWORD,QUANTIFIER,STOP_SENTENCE = range(7)
    DEFAULT = ['dict.dic', 'surname.dic', 'suffix.dic', 'ext_stopword.dic','stopword.dic', 'quantifier.dic']
    DEFAULT_DICT = []

class WordBase(object):
    def __init__(self):
        self.base_freq = 0
        self.base_ps = 0.0
        self.type = ''

class DictBase(object):
    def __init__(self):
        self._data = {}
        self.total = 0

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return (self._data)

    def __getitem__(self, node):
        if self._data.has_key(node):
            return self._data[node]
        else:
            return None
    
    def __iter__(self):
        return self._data.__iter__()

    def iteritems(self):
        return self._data.iteritems()
    
    def add_term(self, term, word):
        self._data[term] = word
        self.total += word.base_freq

    def normalize(self):
        #ignore which the total is zero
        if self.total == 0:
            return
        for term, word in self._data.iteritems():
            word.base_ps = math.log(float(word.base_freq)/self.total)

    def has_key(self, term):
        return self._data.has_key(term)

def __get_sentence_dict():
    cutlist = " .[。，,！……!《》<>\"':：？\?、\|“”‘’；]{}（）{}【】()｛｝（）：？！。，;、~——+％%`:“”＂'‘\n\r"
    if not isinstance(cutlist, unicode):
        try:
            cutlist = cutlist.decode('utf-8')
        except UnicodeDecodeError:
            cutlist = cutlist.decode('gbk','ignore')
    cutlist_dict = {}
    for c in list(cutlist):
        word = WordBase()
        cutlist_dict[c] = word
    return cutlist_dict

def get_dict(type):
    global DICT_INIT
    if type < 0 or type >= DICTS.LEN:
        return {}
    if DICT_INIT:
        return DICTS.DEFAULT_DICT[type]
    with DICT_LOCK:
        if DICT_INIT:
            return DICTS.DEFAULT_DICT[type]
        
        print >> sys.stderr, 'Start load dict...'
        t1 = time.time()
        curpath = os.path.normpath( os.path.join( os.getcwd(), os.path.dirname(__file__) )  )
        tmp_file = os.path.join(tempfile.gettempdir(), "yaha.cache")
        load_ok = True

        # Use pickle to load dict
        if os.path.exists(tmp_file):
            tmp_file_time = os.path.getmtime(tmp_file)
            for p in DICTS.DEFAULT:
                if tmp_file_time < os.path.getmtime(os.path.join(curpath, 'dict', p)):
                    load_ok = False
                    break
            if load_ok:
                try:
                    DICTS.DEFAULT_DICT = pickle.load(open(tmp_file, 'rb'))
                    print >> sys.stderr, 'Load dict from: ' + tmp_file
                except:
                    DICTS.DEFUALT_DICT = []
                    load_ok = False
        else:
            load_ok = False

        if not load_ok:
            for i in xrange(0, DICTS.LEN-1, 1):
                new_dict = DictBase()
                with codecs.open(os.path.join(curpath, 'dict', DICTS.DEFAULT[i]), "r", "utf-8") as file:
                    for line in file:
                        tokens = line.split(" ")
                        term = tokens[0].strip()
                        word = WordBase()
                        if len(tokens) >= 2:
                            word.base_freq = int(tokens[1].strip())
                        if len(tokens) >= 3:
                            word.type = tokens[2].strip()
                        new_dict.add_term(term, word)
                new_dict.normalize()
                DICTS.DEFAULT_DICT.append(new_dict)

            # The sentence_dict is alwayse at last
            DICTS.DEFAULT_DICT.append(__get_sentence_dict())

            # save to tempfile
            tmp_file = os.path.join(tempfile.gettempdir(), "yaha.cache")
            print >> sys.stderr, 'save dicts to tmp_file: ' + tmp_file
            try:
                tmp_suffix = "."+str(random.random())
                with open(tmp_file+tmp_suffix,'wb') as temp_cache_file:
                    pickle.dump(DICTS.DEFAULT_DICT, temp_cache_file, protocol=pickle.HIGHEST_PROTOCOL)
                if os.name=='nt':
                    import shutil
                    replace_file = shutil.move
                else:
                    replace_file = os.rename
                replace_file(tmp_file + tmp_suffix, tmp_file)
            except:
                print >> sys.stderr, "dump temp file failed."
                import traceback
                print >> sys.stderr, traceback.format_exc()

        print >> sys.stderr, 'End load dict ', time.time() - t1, "seconds."
        DICT_INIT = True

    return DICTS.DEFAULT_DICT[type]

class CuttingBase(object):
    def __init__(self):
        self.stage = -1

    #def cut_stage1(self, cuttor, sentence):
    #    pass
    #def cut_stage2(self, cuttor, sentence, graph):
    #    pass
    #def cut_stage3(self, cuttor, sentence, graph, start, end):
    #    pass
    #def cut_stage4(self, sentence, graph, contex):
    #    pass

class RegexCutting(CuttingBase):
    # Not surpport stage 1 now
    def __init__(self, rex, stage=2):
        super(CuttingBase, self).__init__()
        self.stage = 2
        #self.cut_stage2 = self.cut_regex
        self.rex = rex

    def cut_stage2(self, cuttor, sentence, graph):
        for m in self.rex.finditer(sentence):
            graph.add_edge(m.start(0),m.end(0), cuttor.default_prob)
    
class SurnameCutting(CuttingBase):
    def __init__(self):
        super(CuttingBase, self).__init__()
        self.stage = 3
        self.dict = get_dict(DICTS.SURNAME)

    def cut_stage3(self, cuttor, sentence, graph, start, end):
        klen = end-start
        if klen > 4:
            return

        if klen >= 2 and self.dict.has_key(sentence[start:start+1]):
            # TODO for this prob
            graph.add_edge(start,start+2, cuttor.refer_prob*0.8)
        if klen > 2 and self.dict.has_key(sentence[start:start+1]):
            graph.add_edge(start,start+3, cuttor.refer_prob*1.2)
        if klen > 2 and self.dict.has_key(sentence[start:start+2]):
            graph.add_edge(start,start+3, cuttor.refer_prob*1.2)
        if klen > 3 and self.dict.has_key(sentence[start:start+2]):
            graph.add_edge(start,start+4, cuttor.refer_prob*1.5)

class SurnameCutting2(CuttingBase):
    def __init__(self):
        super(CuttingBase, self).__init__()
        self.stage = 4
        self.dict = get_dict(DICTS.SURNAME)
        self.stop_dict = get_dict(DICTS.EXT_STOPWORD)
        
    def cut_stage4(self, cuttor, sentence, graph, contex):
        start = contex['index']
        path = contex['path']
        new_path = contex['new_path']
        n = len(path)
        i = start
        while i < n:
            klen = path[i]-path[i-1]
            #print 'surname', klen, sentence[path[i-1]:path[i]]
            if klen >= 1 and klen <= 2 and self.dict.has_key(sentence[path[i-1]:path[i]]):
                if i < n-2:
                    klen2 = path[i+1]-path[i]
                    klen3 = path[i+2]-path[i+1]
                    if klen2==1 and klen3==1 and not self.stop_dict.has_key(sentence[path[i+1]:path[i+2]]):
                        #get one name
                        new_path.append(path[i+2])
                        i += 3
                        continue
                    if klen2==1:
                        new_path.append(path[i+1])
                        i += 2
                    if klen2==2:
                        new_path.append(path[i+1])
                        i += 2
                        continue
                elif i < n-1:
                    klen2 = path[i+1]-path[i]
                    if klen2 <= 2 and klen2 >=1:
                        new_path.append(path[i+1])
                        i = path[i+1]
                        i += 2
                        continue
                else:
                    return i
            else:
                #next stage to doit
                return i
        return i

class SuffixCutting(CuttingBase):
    def __init__(self):
        super(CuttingBase, self).__init__()
        self.stage = 4
        self.dict = get_dict(DICTS.SUFFIX)
        self.stop_dict = get_dict(DICTS.EXT_STOPWORD)
    
    def cut_stage4(self, cuttor, sentence, graph, contex):
        start = contex['index']
        path = contex['path']
        new_path = contex['new_path']
        n = len(path)
        n2 = len(new_path)
        i = start
        modify = False
        
        #print 'suffix', i, sentence[path[i-1]:path[i]]

        klen = path[i]-path[i-1]
        if klen >= 1 and klen <= 2 and n2 > 2 and self.dict.has_key(sentence[path[i-1]:path[i]]):
            j = n2-1
            while j > 1:
                klen2 = new_path[j]-new_path[j-1]
                if klen2 == 1 and not self.stop_dict.has_key(sentence[new_path[j-1]:new_path[j]]):
                    new_path.pop()
                    contex['single'] -= 1
                    modify = True
                    j -= 1
                    continue
                if klen2 > 1 and cuttor.word_type(sentence[new_path[j-1]:new_path[j]], 'n'):
                    new_path.pop()
                    modify = True
                break
        if modify:
            new_path.append(path[i])
            return i+1
        else:
            return i

#base_class
class BaseCuttor(object):
    def __init__(self):
        #self.WORD_MAX = 8
        # Use 6 is better, but it will not cut the word longer than 5
        self.WORD_MAX = 6
        self.refer_prob = 15.0
        self.default_prob = 150.0
        self.topk = 1
        self.__best_path = self.__cut_graph_simple
        self.stages = []
        self.stages.append([])
        self.stages.append([])
        self.stages.append([])
        self.stages.append([])
        self.sentence_dict = get_dict(DICTS.STOP_SENTENCE)
        self.stage1_regex = None
        self.fn_stage1 = self.__stage1_null

    def exist(self, term):
        pass
    def get_prob(self, term):
        pass
    def word_type(self, term, type):
        return False

    def cut_to_sentence(self, line):
        if not isinstance(line, unicode):
            try:
                line = line.decode('utf-8')
            except UnicodeDecodeError:
                line = line.decode('gbk','ignore')
        str = ''
        for c in line:
            if self.sentence_dict.has_key(c):
                # stage 1
                for s,need_cut in self.fn_stage1(str):
                    yield (s, need_cut)

                str = ''
                yield (c, False)
            else:
                str += c
        if str != '':
            for s,need_cut in self.fn_stage1(str):
                yield (s, need_cut)

    # Only support one regex for stage1 now
    def set_stage1_regex(self, rex):
        if type(rex) == str:
            self.stage1_regex = re.compile(rex, re.I|re.U)
        else:
            self.stage1_regex = rex
        self.fn_stage1 = self.__do_stage1

    def __stage1_null(self, sentence):
        yield (sentence, True)

    def __do_stage1(self, sentence):
        start = 0
        for m in self.stage1_regex.finditer(sentence):
            yield (sentence[start:m.start(0)], True)
            yield (sentence[m.start(0):m.end(0)], False)
            start = m.end(0)
        yield (sentence[start:], True)

    def add_stage(self, obj, stage=-1):
        new_stage = stage
        if new_stage == -1:
            new_stage = obj.stage
        if new_stage < 1 or new_stage > 4:
            return
        self.stages[new_stage-1].append(obj)

    def set_topk(self, topk):
        self.topk = topk
        if topk > 1:
            self.__best_path = self.__cut_graph_topk
        else:
            self.__best_path = self.__cut_graph_simple

    def get_graph(self, sentence):
        n = len(sentence)
        i,j = 0,0
        graph = Graph(n+1, self.default_prob)

        # stage2
        for stage in self.stages[1]:
            stage.cut_stage2(self, sentence, graph)

        while i < n:
            for j in xrange(i, min(n,i+self.WORD_MAX), 1):
                if self.exist(sentence[i:j+1]):
                    graph.add_edge(i,j+1,0-self.get_prob(sentence[i:j+1]))
                else:
                    # stage3
                    for stage in self.stages[2]:
                        stage.cut_stage3(self, sentence, graph, i, j+1)
            i += 1
            j = i
        return graph

    def __new_path(self, sentence, graph, contex):
        path = contex['path']
        # stage4
        index = 1
        n_path = len(path)
        #print item['path']
        while index < n_path:
            i = index
            for stage in self.stages[3]:
                i = stage.cut_stage4(self, sentence, graph, contex)
                if i >= n_path:
                    break
                #print 'index', i
                contex['index'] = i
            if i >= n_path:
                break
            if path[i]-path[i-1] == 1:
                contex['single'] += 1
            contex['new_path'].append(path[i])
            index = i+1
            contex['index'] = index

    def __cut_graph_simple(self, sentence, graph):
        (_, path) = quick_shortest(graph)
        contex = {'path':path,'new_path':[0], 'single':0, 'index':1}
        self.__new_path(sentence, graph, contex=contex)
        return contex['new_path']

    def choise_best(self, sentence, graph, items):
        paths = []
        for item in items:
            path = item['path']
            contex = {'path':path,'new_path':[0], 'single':0, 'index':1}
            self.__new_path(sentence, graph, contex=contex)
            paths.append(contex)
        sorted(paths, key=lambda p:p['single'])
        return paths[0]['new_path']

    def __cut_graph_topk(self, sentence, graph):
        items = ksp_yen(graph, 0, graph.N-1, self.topk)
        return self.choise_best(sentence, graph, items)
    
    def __cut_graph(self, sentence):
        graph = self.get_graph(sentence)
        path = self.__best_path(sentence, graph)
        
        for i in xrange(1, len(path), 1):
            yield sentence[path[i-1]:path[i]]

    def __cut_all(self, sentence):
        graph = self.get_graph(sentence)

        old_j = -1
        for k,v in graph.iteritems():
            if len(v) == 1 and k > old_j:
                yield sentence[k:k+1]
            else:
                for j,_p in v.iteritems():
                    if j > k+1:
                        yield sentence[k:j]
                        old_j = j-1

    def cut_topk(self, sentence, topk):
        if not isinstance(sentence, unicode):
            try:
                sentence = sentence.decode('utf-8')
            except UnicodeDecodeError:
                sentence = sentence.decode('gbk','ignore')

        graph = self.get_graph(sentence)
        items = ksp_yen(graph, 0, graph.N-1, topk)

        for item in items:
            words = []
            path = item['path']
            for i in xrange(1, len(path), 1):
                words.append(sentence[path[i-1]:path[i]])
            yield words

    def cut(self, sentence):
        for s,need_cut in self.cut_to_sentence(sentence):
            if s == '':
                continue
            elif need_cut:
                for word in self.__cut_graph(s):
                    yield word
            else:
                yield s

    def cut_all(self, sentence):
        for s,need_cut in self.cut_to_sentence(sentence):
            if s == '':
                continue
            elif need_cut:
                for word in self.__cut_all(s):
                    yield word
            else:
                yield s
    
    def tokenize(self, unicode_sentence, search = False):
        if not isinstance(unicode_sentence, unicode):
            raise Exception("Yaha: the input parameter should be unicode.")
        start = 0
        if search:
            for term in self.cut(unicode_sentence):
                width = len(term)
                if width > 2:
                    for i in xrange(width - 1):
                        gram2 = term[i:i+2]
                        if self.exist(gram2):
                            yield (gram2,start+i,start+i+2)
                if width > 3:
                    for i in xrange(width - 2):
                        gram3 = term[i:i+3]
                        if self.exist(gram3):
                            yield (gram3, start+i, start+i+3)
                yield (term, start, start+width)
                start += width
        else:
            for term in self.cut(unicode_sentence):
                width = len(term)
                yield (term, start, start+width)
                start += width

class Cuttor(BaseCuttor):
    
    def __init__(self):
        super(Cuttor, self).__init__()
        
        #self.set_topk(3)
        self.dict = get_dict(DICTS.MAIN)

        self.refer_prob =  (0-math.log(3.0/self.dict.total))
        self.default_prob = 10.0*self.refer_prob

    def exist(self, term):
        word = self.dict[term]
        if word and word.base_freq > 0:
            return True
        return False
    
    def word_type(self, term, type):
        word = self.dict[term]
        if word:
            return (word.type.find(type)>=0)
        return False

    def get_prob(self, term):
        word = self.dict[term]
        if word:
            return word.base_ps
        else:
            return (0-self.default_prob)




from distutils.core import setup  
setup(name='yaha',
      version='0.01.alpha',  
      keywords=('word', 'segmenetation', 'keyword', 'summerize'),
      description='Chinese Words Segementation Utilities',
      license = 'MIT License',
      author='Janson, Yuzheng',  
      author_email='gandancing@gmail.com',  
      url='http://github.com/jannson/yaha',  
      packages=['yaha'],  
      package_dir={'yaha':'yaha'},
      package_data={'yaha':['*.*','analyse/*','dict/*']}
)
